module Test.Data.String.Search.KarpRabin.Spec where

import Data.Array (index)
import Data.Foldable (all)
import Data.Maybe (Maybe(..), isJust)
import Data.String (drop, stripPrefix, Pattern(..))
import Data.String.Search.KarpRabin (indicesOfAny, mkRollingHash, hashU64, rehashU64, hashStruct, fromCodePoint)
import Data.String.Search.Utils (slidingWindow)
import Data.Tuple (Tuple(..))
import Data.UInt64 (unsafeFromInt)
import Prelude
import Test.Data.String.Search.Utils (CodePointA(..))
import Test.QuickCheck ((<?>))
import Test.Spec (Spec, describe, it)
import Test.Spec.Assertions (shouldEqual)
import Test.Spec.QuickCheck (quickCheck')



validIndices :: Array String -> String -> Boolean
validIndices pats input = all validIndex (indicesOfAny pats input)
  where
    validIndex (Tuple i ps) = all validPat ps
      where
        input' = drop i input
        validPat p =
          case index pats p of
            Just pat -> isJust (stripPrefix (Pattern pat) input')
                        -- <?> (show input' <> " should start with " <> show pat)
            Nothing  -> false -- Failed "out of bounds pattern"

spec :: Spec Unit
spec =
  describe "KarpRabin" do
    it "sliding window works" do
      slidingWindow [1, 2, 3, 4] 2 `shouldEqual`
        [[1, 2], [2, 3], [3, 4]]
    
    it "rolling hash works 1" do
      let rh = mkRollingHash (unsafeFromInt 7) 3
      let a = unsafeFromInt 1
      let b = unsafeFromInt 2
      let c = unsafeFromInt 3
      let d = unsafeFromInt 4

      let h1 = hashU64 rh [a, b, c]
          h2 = hashU64 rh [b, c, d]
          h3 = hashU64 rh [c, d, a]

      h2 `shouldEqual` rehashU64 rh h1 a d
      h3 `shouldEqual` rehashU64 rh h2 b a

    it "rolling hash works 2 (quickcheck)" $ do
      let rh = mkRollingHash (unsafeFromInt 256) 3

      quickCheck' 2000 \(CodePointA a') (CodePointA b') (CodePointA c') (CodePointA d') ->
        let a = fromCodePoint a'
            b = fromCodePoint b'
            c = fromCodePoint c'
            d = fromCodePoint d'
            h1 = hashU64 rh [a, b, c]
            h2 = hashU64 rh [b, c, d]
        in
        h2 == rehashU64 rh h1 a d
          <?> ( "Fail for: " <> show [a', b', c', d']
               <> ", h1 = " <> show h1
               <> ", h2 = " <> show h2
              <> ", rehash = " <> show (rehashU64 rh h1 a d))
    
    it "works on a single pattern matching two times" do
      let pats   = ["ab"]
      let input  = "abcbab"
      let output = [Tuple 0 [0], Tuple 4 [0]]
      indicesOfAny pats input `shouldEqual` output

    it "works on a many unmatching patterns" do
      let pats   = ["abd","e","bac","abcbabe"]
      let input  = "abcbab"
      let output = []
      indicesOfAny pats input `shouldEqual` output

    it "works when pattern is in the middle" do
      let pats = ["abd"]
      let input = "eeabdee"
      let output = [Tuple 2 [0]]
      let hs = hashStruct pats
      let h1 = hs.hash "abd"
      let h2 = hs.hash "bde"
      h2 `shouldEqual` (hs.rehashChar h1 'a' 'e')
      indicesOfAny pats input `shouldEqual` output

    it "works when pattern is in the middle, with space in it (ps #598)" do
      let pats = [" abd"]
      let input = "e abdee"
      let output = [Tuple 1 [0]]
      let hs = hashStruct pats
      let h1 = hs.hash "e ab"
      let h2 = hs.hash " abd"
      h2 `shouldEqual` (hs.rehashChar h1 'e' 'd')
      indicesOfAny pats input `shouldEqual` output

    it "works on a bit more complex case" do
      let pats   = ["ab","cb","bc","bca"]
      let input  = "abcbab"
      let output = [Tuple 0 [0]
                   ,Tuple 1 [2]
                   ,Tuple 2 [1]
                   ,Tuple 4 [0]
                   ]
      indicesOfAny pats input `shouldEqual` output

    it "works with overlaps" do
      let pats   = ["aba"]
      let input  = "ababa"
      let output = [Tuple 0 [0]
                   ,Tuple 2 [0]
                   ]
      indicesOfAny pats input `shouldEqual` output

    it "returns valid indices" do
      validIndices ["a","ab","ba","abc","aba","abab","abcde"]
                   "ababarbabacbbababcaccacabbababa"
          `shouldEqual` true

    it "returns valid indices 2000 random samples" do
      quickCheck' 2000 validIndices

    it "works for purescript-gargantext issue #598, test 1" do
      let pats = [" conception " ]
      let input = " passant  d'une  conception  quasi  solipsiste  ou  monadique  de  la  sensation "
      let output = [Tuple 16 [0]]
      indicesOfAny pats input `shouldEqual` output

    it "works for purescript-gargantext issue #598, test 2" do
      let pats = [" conception ", " notion "]
      let input = " passant  d'une  conception  quasi  solipsiste  ou  monadique  de  la  sensation "
      let output = [Tuple 16 [0]]
      indicesOfAny pats input `shouldEqual` output
